#!/usr/local/bin/perl
# get_db_seq_table_input.pl
# Sheila Podell
# March 23, 2009

# prepares input for darkhorse db_seq_ids table
# DarkHorse database must already have two tables loaded 

	# gi_tax_prot
	# precalc_lineages
	
# input is tab-delimited (nr.info) extracted from fasta file of Genbank nr amino acid seqs
# produced by get_nr_seq_info.pl. Fields:

	# full_sequence_id	  gi_num  seq_length(aa)   annotation

# output (STDOUT) is a tab-delimited list of all nr sequences, that combines nr.info with lineages
# lineages are obtained from pre-existing MySQL table
	
use warnings; 
use strict;
use Getopt::Long;
use DBI;

# get command line arguments, defaults 
	my ($db, $datafile, $sth, $sql, $config_filename, $query, @row);
	my $USAGE;
	my $message = qq(
	Usage: $0  -c config filename -i nr_seq_info file )."\n\n";
	GetOptions( 
			    "c=s" => \$config_filename,
			    "i=s" => \$datafile,
			    
				);	     
	
	if($USAGE || !$config_filename || !$datafile) 
	{
		print STDERR $message;
		exit(0);
	} 
	
	unless (-s $config_filename)
	{
		print STDERR "  Couldn't find config file $config_filename \n";
		exit(0);
	}
	unless (-s $datafile)
	{
		print STDERR "  Couldn't find config file $datafile \n";
		exit(0);
	}	

# get SQL database parameters from config file
    my %config_params= &read_config_file($config_filename);
	my $db_name = $config_params{db_name} || "not found";
	my $db_program = $config_params{db_program} || "not found";
	my $db_host = $config_params{db_host} || "not found";
	my $db_user = $config_params{db_user} || "not found";
	my $db_user_password = $config_params{db_user_password} || "not found";
	my $max_lines_per_packet = $config_params{max_lines_per_packet} || "not found";
		
	unless (-s "$datafile")
	{
		print STDERR "  Couldn't find nr.info file $datafile \n";
		exit(0);
	}
	
# connect to database
	my $db_path = "DBI:$db_program:$db_name:$db_host:";
	my $dbh = DBI->connect($db_path, $db_user, $db_user_password) or die "Cannot connect: $DBI::errstr"; 	

# check to make sure gi_tax_prot table is available for cross-reference
	my $tablename1 = "gi_tax_prot";
	my $found_table1 = &check_table("$tablename1");
	unless ($found_table1 ==1)
	{	
		print STDERR "  Couldn't find table $tablename1\n";
		print STDERR "  Required to load table $tablename1.\n";
		exit(0);
	}
	
# check to make sure precalc_lineages table is available for cross-reference
	my $tablename2 = "precalc_lineages";
	my $found_table2 = &check_table("$tablename2");
	unless ($found_table2 ==1)
	{	
		print STDERR "  Couldn't find table $tablename2\n";
		print STDERR "  Required to load table $tablename2.\n";
		exit(0);
	}

# go through each line of input file (once) to merge with species (tax_id) info from existing db
# Can't put whole input file into a hash at once (it's too big), so do it in sections
# $max_section_lines is an arbitrary number of lines, small enough to avoid exceeding avail. memory

	my %seqinfo = ();	# key = gi number value = tab-delimited string of info for that sequence
	my $section_num = 1;
	my $section_lines = 0;
	my $max_section_lines = 100000;	
	my $outfile_name = "db_seq_ids_input";
	open (INFILE, "$datafile") or die "couldn't open $datafile, $!\n";	 
	while (<INFILE>)
	{
		# parse line
			chomp;
			next if ($_ =~ /^\s+$/);
			my @dataline = split "\t", $_;
			next unless (scalar @dataline >3);
			$section_lines++;			
			my $gi_num = $dataline[1];
	
	# collect info in temporary hash
		$seqinfo{$gi_num} = "$_";

	# when number of hash entries fills sections, process and clear out memory
		if ($section_lines % $max_section_lines == 0 || eof)
		{
			my @gi_list = (keys %seqinfo);
			my $num_to_get = scalar @gi_list;
			print STDERR "  processing seq data group $section_num, line $.\n";	
			
	 		$query = qq(select gi_num, gi_tax_prot.ncbi_tax_id, species_name
 		from gi_tax_prot, precalc_lineages
 		where gi_tax_prot.ncbi_tax_id = precalc_lineages.ncbi_tax_id
 		and gi_num
 		);
 		$query .= "in (";
 		my @output = &sql_select_list(\@gi_list, $query);
 		
 	# need to identify entries where tax_id is missing from pre_calc lineage list
 	# happens when gi_taxid_prot entry for tax_id is missing from ncbi names file (ncbi mistakes)
 	# there are a lot of these missing tax_id's (thousands), presumably obsoleted					
 		
		# print output, joining together sql query with nr.info fields
		# keep track of found_ids so the ones not found in precalc_lineage table can be added later
		 my %sql_found_ids =();	# key = gi_num	
		 foreach my $next (@output)
			{
				my @next = split "\t", $next;
				my $db_gi = $next[0];
				my $tax_id = $next[1];
				my $species_name = $next[2];
				$sql_found_ids{$db_gi}++;
				
			unless (defined $seqinfo{$db_gi})
			{
				print STDERR "next_line=$next\n";
				print STDERR "undefined gi num for $next ($db_gi)\n";
				exit(0);
			}
				my @nr_info = split "\t", $seqinfo{$db_gi};
				my $full_id = $nr_info[0];
				my $nr_gi = $nr_info[1];
				my $seq_len = $nr_info[2] || "seq_len";
				my $annot = $nr_info[3];
				my $alt_id = "alt_id";	# should parse this from full
				my @id_sections = split '\|', "$full_id";
				if (defined $id_sections[3])
				{
					$alt_id = $id_sections[3];
				}
				my @seq_fields = (			
					$full_id,
					$nr_gi,
					$alt_id,
					$tax_id,
					$species_name,
					$seq_len,
					$annot,
					);
				my $seqstring = join "\t", @seq_fields;
				print "$seqstring\n";
	 		};
	 		
	 		foreach my $key (keys %seqinfo)
	 		{
	 	# add back nr sequences that the SQL query missed because of bad tax_id, missing precalc_lineage
	 			unless (exists $sql_found_ids{$key})
	 			{	 				
	 				my @tmp = split "\t", $seqinfo{$key};
	 				my $full_id = $tmp[0];
	 				my $nr_gi = $tmp[1];
	 				my $alt_id = "";
				 	my @id_sections = split '\|', "$full_id";
					if (defined $id_sections[3])
					{
						$alt_id = $id_sections[3];
					}
	 				my $tax_id = 0;
	 				my $species_name = "unknown";
	 				
	 				my $seq_len = $tmp[2];
	 				my $annot = $tmp[3];

	 				my @seq_fields = (			
					$full_id,
					$nr_gi,
					$alt_id,
					$tax_id,
					$species_name,
					$seq_len,
					$annot,
					);
	 				my $seqstring = join "\t", @seq_fields;
					print "$seqstring\n";
	 			}	 		
	 		}
	 						
		# reset counters, clear temporary memory
			$section_num++;
			$section_lines = 0;
			%seqinfo = ();
			%sql_found_ids = ();
		}			
	}
	close INFILE;
	
	
#########################################################
# SUBROUTINES
#########################################################
sub read_config_file
{
    my ($filename) = @_;
    open (INPUT, $filename)  or die "can't open DarkHorse config file $filename\n$!\n";
	my %dh_config = ();
	while (<INPUT>)
	{
		next if ($_ =~ /^\s+$/);
		next if ($_ =~ /^#/);
		if ($_ =~ /\[(.+)\]=\s*(.+)/)
		{
			$dh_config{$1}=$2;		
		}			
	}
	close INPUT;			
    return %dh_config;
}

sub check_table
{
	my ($tablename) = @_;
	my $found_table = 0;
	$sql = "show tables;";
	$sth = $dbh->prepare($sql)
    	or dieStackTrace("Cannot prepare SELECT '$sql':\n<b>$DBI::errstr</b>");
	$sth->execute()
    	or dieStackTrace("Cannot execute SELECT '$sql':\n<b>$DBI::errstr</b>");
	my @row = ();
	my $element = ();
	while (@row = $sth->fetchrow_array)
	{
		foreach $element (@row)
		{
			if ($element =~/$tablename/)
			{
				$found_table = 1;
				last;
			}
		}
	}
	return $found_table;
}
sub sql_select_list
{		
	my ($list, $firstline) = @_;
	my @select_list = @$list;
	my $total_num = scalar @select_list;
	my $sql = $firstline;
	
	my @results_tbl = ();
		
	unless ($total_num >0)
	{
		print STDERR "couldn't find any selections in query list\n";
		exit(0);
	}
	my $count = 0;
		
	foreach (@select_list)
	{
		$count++;
		chomp;
		$_ =~ s/'//g;		# get rid of internal quotes
		$sql .= "'$_',";
	# send large select statements in pieces to avoid exceeding MySql max_allowed_packet size	
		if ($count % $max_lines_per_packet == 0)
		{
			$sql = substr($sql, 0, -1); # get rid of extra newline and comma, if present
			$sql .= ");\n";
						
			$sth = $dbh->prepare($sql)
    			or dieStackTrace("Cannot prepare SELECT '$sql':\n<b>$DBI::errstr</b>");
			$sth->execute()
    			or dieStackTrace("Cannot execute SELECT '$sql':\n<b>$DBI::errstr</b>");
			@row = ();
			while (@row = $sth->fetchrow_array)
			{				
				my $row_txt;
				foreach my $element (@row)
				{
					$row_txt .= "$element\t";
				}
				push @results_tbl, "$row_txt";
			}
			$sql = $firstline;			
		}
		elsif ($count == $total_num)
		{
			$sql = substr($sql, 0, -1); # get rid of extra newline and comma, if present
			$sql .= ");\n";

			$sth = $dbh->prepare($sql)
    			or dieStackTrace("Cannot prepare SELECT '$sql':\n<b>$DBI::errstr</b>");
			$sth->execute()
    			or dieStackTrace("Cannot execute SELECT '$sql':\n<b>$DBI::errstr</b>");
			@row = ();
			while (@row = $sth->fetchrow_array)
			{
		 		my $row_txt;
				foreach my $element (@row)
				{
					$row_txt .= "$element\t";
				}
				push @results_tbl, "$row_txt";
			}
		}
	}
	
	return @results_tbl;
}

sub sql_select
{		
	my ($list, $table_name, $firstline) = @_;
	my @select_list = @$list;
	my $total_num = scalar @select_list;
	my $sql = $firstline;
	
	my @results_tbl = ();
		
	unless ($total_num >0)
	{
		print STDERR "couldn't find any selections in list for $table_name\n";
		exit(0);
	}
	my $count = 0;			
	foreach (@select_list)
	{
		$count++;
		chomp;
		$_ =~ s/'//g;		# get rid of internal quotes
		$sql .= "'$_',";
	# send large select statements in pieces to avoid exceeding MySql max_allowed_packet size	
		if ($count % $max_lines_per_packet == 0)
		{
			$sql = substr($sql, 0, -1); # get rid of extra newline and comma, if present
			$sql .= ");\n";
						
			$sth = $dbh->prepare($sql)
    			or dieStackTrace("Cannot prepare SELECT '$sql':\n<b>$DBI::errstr</b>");
			$sth->execute()
    			or dieStackTrace("Cannot execute SELECT '$sql':\n<b>$DBI::errstr</b>");
			@row = ();
			while (@row = $sth->fetchrow_array)
			{				
				my $row_txt;
				foreach my $element (@row)
				{
					$row_txt .= "$element\t";
				}
				push @results_tbl, "$row_txt";
			}
			$sql = $firstline;
		}
		elsif ($count == $total_num)
		{
			$sql = substr($sql, 0, -1); # get rid of extra newline and comma, if present
			$sql .= ");\n";
			$sth = $dbh->prepare($sql)
    			or dieStackTrace("Cannot prepare SELECT '$sql':\n<b>$DBI::errstr</b>");
			$sth->execute()
    			or dieStackTrace("Cannot execute SELECT '$sql':\n<b>$DBI::errstr</b>");
			@row = ();
			while (@row = $sth->fetchrow_array)
			{
		 		my $row_txt;
				foreach my $element (@row)
				{
					$row_txt .= "$element\t";
				}
				push @results_tbl, "$row_txt";
			}
		}
	}	
	return @results_tbl;
}


