#!/usr/local/bin/perl
# get_nr_species_list.pl
# Sheila Podell
# April 1, 2009

# takes as input an ncbi file gi_taxid_prot.dmp and a darkhorse database config file from
#	ftp://ftp.ncbi.nih.gov/pub/taxonomy/gi_taxid_prot.dmp.gz

# NCBI taxonomy names and nodes fields must already be loaded into the darkhorse database
#	ftp://ftp.ncbi.nih.gov//pub/taxonomy/taxdump.tar.gz 

# output is a list of species names for all tax_ids found in nr, named "species_list"

use warnings; 
use strict;
use Getopt::Long;
use DBI;

# get command line arguments, defaults 
	my ($db, $datafile, $sth, $sql, $config_filename, $query, @row);
	my $USAGE;
	my $message = qq(
	Usage: $0  -c config filename  )."\n\n";
	GetOptions( 
			    "c=s" => \$config_filename,
				);	     
	
	if($USAGE || !$config_filename) 
	{
		print STDERR $message;
		exit(0);
	} 
	
	unless (-s $config_filename)
	{
		print STDERR "  Couldn't find config file $config_filename \n";
		exit(0);
	}	

# get SQL database parameters from config file
    my %config_params= &read_config_file($config_filename);
	my $db_name = $config_params{db_name} || "not found";
	my $db_program = $config_params{db_program} || "not found";
	my $db_host = $config_params{db_host} || "not found";
	my $db_user = $config_params{db_user} || "not found";
	my $db_user_password = $config_params{db_user_password} || "not found";
	my $max_lines_per_packet = $config_params{max_lines_per_packet} || "not found";
	my $gi_taxid_prot_index = $config_params{gi_taxid_prot_dmp_path} || "not found";
	
	unless (-s "$gi_taxid_prot_index")
	{
		print STDERR "  Couldn't find gi_taxid_prot.dmp file $gi_taxid_prot_index \n";
		exit(0);
	}
	
# connect to database
	my $db_path = "DBI:$db_program:$db_name:$db_host:";
	my $dbh = DBI->connect($db_path, $db_user, $db_user_password) or die "Cannot connect: $DBI::errstr"; 	

# check to make sure names table is available for cross-reference
	my $tablename1 = "names";
	my $found_table1 = &check_table("$tablename1");
	unless ($found_table1 ==1)
	{	
		print STDERR "  Couldn't find table $tablename1\n";
		print STDERR "  Required to load table $tablename1.\n";
		exit(0);
	}
	
# check to make sure nodes table is available for cross-reference
	my $tablename2 = "nodes";
	my $found_table2 = &check_table("$tablename2");
	unless ($found_table2 ==1)
	{	
		print STDERR "  Couldn't find table $tablename2\n";
		print STDERR "  Required to load table $tablename2.\n";
		exit(0);
	}

# get list of unique tax id numbers from the input directory 
	my $id_lines = `cut -f2 $gi_taxid_prot_index |sort |uniq`;
	my @taxidlist = split "\n", $id_lines;
	foreach (@taxidlist)
	{
		chomp $_;
	}
	my $num_species = scalar @taxidlist;
	print STDERR "$num_species tax id numbers found\n";
	unless (scalar @taxidlist >0)
	{
		exit(0);
	}

# get list of species names associated with these tax ids
	$query = qq(select name_txt
	from names
	where name_class = "scientific name"
	and tax_id
	);
	$query .= "in (";
	my @output = &sql_select_list(\@taxidlist, $query);
	
# check output for exact match species names with > 1 ncbi_tax_id
# (shouldn't be, but is sometimes true	

	my $num_names = scalar @output;
	print STDERR "$num_names species names found\n";
	my %uniq_names = (); # key = scientific name # value = # tax_id lines
	foreach (@output)
	{
		$uniq_names{$_}++;
		if ($uniq_names{$_} > 1)
		{
			#print STDERR "WARNING: $_ has more than one tax_id\n";
		}
		print "$_\n";				
	}
	
#########################################################
# SUBROUTINES
#########################################################
sub read_config_file
{
    my ($filename) = @_;
    open (INPUT, $filename)  or die "can't open DarkHorse config file $filename\n$!\n";
	my %dh_config = ();
	while (<INPUT>)
	{
		next if ($_ =~ /^\s+$/);
		next if ($_ =~ /^#/);
		if ($_ =~ /\[(.+)\]=\s*(.+)/)
		{
			$dh_config{$1}=$2;		
		}			
	}
	close INPUT;			
    return %dh_config;
}

sub check_table
{
	my ($tablename) = @_;
	my $found_table = 0;
	$sql = "show tables;";
	$sth = $dbh->prepare($sql)
    	or dieStackTrace("Cannot prepare SELECT '$sql':\n<b>$DBI::errstr</b>");
	$sth->execute()
    	or dieStackTrace("Cannot execute SELECT '$sql':\n<b>$DBI::errstr</b>");
	my @row = ();
	my $element = ();
	while (@row = $sth->fetchrow_array)
	{
		foreach $element (@row)
		{
			if ($element =~/$tablename/)
			{
				$found_table = 1;
				last;
			}
		}
	}
	return $found_table;
}
sub sql_select_list
{		
	my ($list, $firstline) = @_;
	my @select_list = @$list;
	my $total_num = scalar @select_list;
	my $sql = $firstline;
	
	my @results_tbl = ();
		
	unless ($total_num >0)
	{
		print STDERR "couldn't find any selections in query list\n";
		exit(0);
	}
	my $count = 0;
		
	foreach (@select_list)
	{
		$count++;
		chomp;
		$_ =~ s/'//g;		# get rid of internal quotes
		$sql .= "'$_',";
	# send large select statements in pieces to avoid exceeding MySql max_allowed_packet size	
		if ($count % $max_lines_per_packet == 0)
		{
			$sql = substr($sql, 0, -1); # get rid of extra newline and comma, if present
			$sql .= ");\n";
						
			$sth = $dbh->prepare($sql)
    			or dieStackTrace("Cannot prepare SELECT '$sql':\n<b>$DBI::errstr</b>");
			$sth->execute()
    			or dieStackTrace("Cannot execute SELECT '$sql':\n<b>$DBI::errstr</b>");
			@row = ();
			while (@row = $sth->fetchrow_array)
			{				
				my $row_txt;
				foreach my $element (@row)
				{
					$row_txt .= "$element\t";
				}
				push @results_tbl, "$row_txt";
			}
			$sql = $firstline;			
		}
		elsif ($count == $total_num)
		{
			$sql = substr($sql, 0, -1); # get rid of extra newline and comma, if present
			$sql .= ");\n";

			$sth = $dbh->prepare($sql)
    			or dieStackTrace("Cannot prepare SELECT '$sql':\n<b>$DBI::errstr</b>");
			$sth->execute()
    			or dieStackTrace("Cannot execute SELECT '$sql':\n<b>$DBI::errstr</b>");
			@row = ();
			while (@row = $sth->fetchrow_array)
			{
		 		my $row_txt;
				foreach my $element (@row)
				{
					$row_txt .= "$element\t";
				}
				push @results_tbl, "$row_txt";
			}
		}
	}
	
	return @results_tbl;
}

sub sql_select
{		
	my ($list, $table_name, $firstline) = @_;
	my @select_list = @$list;
	my $total_num = scalar @select_list;
	my $sql = $firstline;
	
	my @results_tbl = ();
		
	unless ($total_num >0)
	{
		print STDERR "couldn't find any selections in list for $table_name\n";
		exit(0);
	}
	my $count = 0;			
	foreach (@select_list)
	{
		$count++;
		chomp;
		$_ =~ s/'//g;		# get rid of internal quotes
		$sql .= "'$_',";
	# send large select statements in pieces to avoid exceeding MySql max_allowed_packet size	
		if ($count % $max_lines_per_packet == 0)
		{
			$sql = substr($sql, 0, -1); # get rid of extra newline and comma, if present
			$sql .= ");\n";
						
			$sth = $dbh->prepare($sql)
    			or dieStackTrace("Cannot prepare SELECT '$sql':\n<b>$DBI::errstr</b>");
			$sth->execute()
    			or dieStackTrace("Cannot execute SELECT '$sql':\n<b>$DBI::errstr</b>");
			@row = ();
			while (@row = $sth->fetchrow_array)
			{				
				my $row_txt;
				foreach my $element (@row)
				{
					$row_txt .= "$element\t";
				}
				push @results_tbl, "$row_txt";
			}
			$sql = $firstline;
		}
		elsif ($count == $total_num)
		{
			$sql = substr($sql, 0, -1); # get rid of extra newline and comma, if present
			$sql .= ");\n";
			$sth = $dbh->prepare($sql)
    			or dieStackTrace("Cannot prepare SELECT '$sql':\n<b>$DBI::errstr</b>");
			$sth->execute()
    			or dieStackTrace("Cannot execute SELECT '$sql':\n<b>$DBI::errstr</b>");
			@row = ();
			while (@row = $sth->fetchrow_array)
			{
		 		my $row_txt;
				foreach my $element (@row)
				{
					$row_txt .= "$element\t";
				}
				push @results_tbl, "$row_txt";
			}
		}
	}	
	return @results_tbl;
}


