#!/usr/local/bin/perl
# recursive_taxonomy.pl
# Sheila Podell
# August 8, 2007

# recursively parses Taxonomy database to get lineage for a species
# takes as input, file containing list of species names, one per line
# outputs semicolon-delimited lineage for each entry to STDOUT

# tab output parameters:

# species_name tax_id lineage

use warnings;
use Getopt::Long;
use DBI;

# global variables
	my $recursion_max = 35;
	my $recursion_count = 0;
	
# get command line arguments, defaults 
	my ($db, $datafile, $sth, $sql, $config_filename);
	my $USAGE;
	my $message = qq(
	Usage: $0 -d database_name -i source_datafile -c config filename <-o allow overwrites> )."\n\n";
	GetOptions( "d=s" => \$db,
			    "i=s" => \$datafile,
			    "c=s" => \$config_filename,
				);	     
	
	if($USAGE || !$db || !$datafile) 
	{
		print STDERR $message;
		exit(0);
	} 
	unless (-s $datafile)
	{
		print STDERR "  Couldn't find source datafile $datafile \n";
		exit(0);
	}
	unless (-s $config_filename)
	{
		print STDERR "  Couldn't find config file $config_filename \n";
		exit(0);
	}	
		
# get SQL database parameters from config file
    my %config_params= &read_config_file($config_filename);
	my $db_name = $config_params{db_name} || "not found";
	my $db_program = $config_params{db_program} || "not found";
	my $db_host = $config_params{db_host} || "not found";
	my $db_user = $config_params{db_user} || "not found";
	my $db_user_password = $config_params{db_user_password} || "not found";
	
# connect to database
	my $db_path = "DBI:$db_program:$db_name:$db_host:";
	my $dbh = DBI->connect($db_path, $db_user, $db_user_password) or die "Cannot connect: $DBI::errstr"; 	

# set up output order for hierarchical keys - stop after genus
	my @output_list = (
		#"no rank",
		"superkingdom",
		"kingdom",
		"superphylum",
		"phylum",
		"subphylum",
		"superclass",
		"class",
		"subclass",
		"infraclass",
		"superorder",
		"order",
		"parvorder",
		"suborder",
		"infraorder",
		"superfamily",
		"family",
		"subfamily",
		"tribe",
		"subtribe",
		"genus",
		"subgenus",
		#"species group",
		#"species subgroup",
		#"species",
		#"subspecies",
		#"forma",
		#"varietas",
		);
	my %output_order =();
	my $index = 0;
	foreach (@output_list)
	{
		$output_order{$_} = $index;
		$index++;
	}
	
# process data	
	open(INPUT,"$datafile")|| die "cannot find input data file $datafile\n";
	while($line=<INPUT>)
	{
        my @rows= ();
        my $num = 0;
        my %query = ();
        my %Tax = ();
        next if($line=~/^#/);
        %Tax=();	#key = order retrieved, value = tax_id
        chomp $line;
    # remove anything with quotes in it
        $line =~ s/'//g;
    # make sure there are no extra spaces at end of line
        if ($line =~ /(.+)\s+$/)
        {
            $line = $1;
        }
        
        my $orig_line = $line;
	
    $sql = "select tax_id from names where name_txt like '$line%'";
	$sth = $dbh->prepare($sql)
    	or dieStackTrace("Cannot prepare SELECT '$sql':\n<b>$DBI::errstr</b>");
	$sth->execute()
    	or dieStackTrace("Cannot execute SELECT '$sql':\n<b>$DBI::errstr</b>");
    $num = $sth->rows;	
	
# if nothing was returned, try query again using first two words
	my $alt_name = "";
	my @words = split /\s/, $line;
	if($num==0)
	{		
		if (defined $words[1])
		{
			$alt_name = "$words[0] "."$words[1]";
		}
        $sql = "select tax_id from names where name_txt like '$alt_name%'";
        $sth = $dbh->prepare($sql)
            or dieStackTrace("Cannot prepare SELECT '$sql':\n<b>$DBI::errstr</b>");
        $sth->execute()
            or dieStackTrace("Cannot execute SELECT '$sql':\n<b>$DBI::errstr</b>");
        $num = $sth->rows;
	}
	
	
# if nothing was returned, try query again using just first word
	if($num==0)
	{
		$words[0] =~ s/\s//g;
		if (defined $words[0])
		{
			$alt_name = "$words[0]";
		}
		$sql = "select tax_id from names where name_txt like '$alt_name%'";
        $sth = $dbh->prepare($sql)
            or dieStackTrace("Cannot prepare SELECT '$sql':\n<b>$DBI::errstr</b>");
        $sth->execute()
            or dieStackTrace("Cannot execute SELECT '$sql':\n<b>$DBI::errstr</b>");
        $num = $sth->rows;
	}
# if still nothing was returned, send message to STDOUT
	if($num==0)
	{
		print STDERR "No tax_id found for $orig_line \n";
	}
	

# get rest of lineage by recursion, or give up and go on to next line
	if($num!=0)
	{
		%query = %{$sth->fetchrow_hashref};			
		$tax_id=$query{tax_id};
		%Tax=&nodes($tax_id,%Tax);
	
#	debug
		foreach (sort {$Tax{$b} <=> $Tax{$a}} keys %Tax)
		{
			#print STDERR "$Tax{$_},\n"; #tax_id numbers only
		}

# output results in reverse recursion order, removing extraneous stuff		
		my $out_line = "$orig_line\t$query{tax_id}\t";
		foreach my $key (sort {$b <=> $a} (keys(%Tax)))
		{
                $sql = "select name_txt from names where tax_id=$Tax{$key} and name_class like 'scientific name'";
                $sth = $dbh->prepare($sql)
                or dieStackTrace("Cannot prepare SELECT '$sql':\n<b>$DBI::errstr</b>");
                $sth->execute()
                or dieStackTrace("Cannot execute SELECT '$sql':\n<b>$DBI::errstr</b>");
                %query = %{$sth->fetchrow_hashref};
				
			# if anything was returned, check to be sure it's rank is approved
				my $current_name = $query{name_txt};				
				$sql = "select rank from nodes where tax_id=$Tax{$key}";
                $sth = $dbh->prepare($sql)
                or dieStackTrace("Cannot prepare SELECT '$sql':\n<b>$DBI::errstr</b>");
                $sth->execute()
                or dieStackTrace("Cannot execute SELECT '$sql':\n<b>$DBI::errstr</b>");
                %query = %{$sth->fetchrow_hashref};
                
				my $current_rank = $query{rank};
				if (exists $output_order{$current_rank} || $current_name =~/Viruses/)
				{
					#get rid of non-informative names	
						$current_name =~ s/unclassified//;
						$current_name =~ s/\(miscellaneous\)//;
						next if  $current_name =~ /candidate\s/;
						next if  $current_name =~ /\(class\)/;
						next if  $current_name =~ /group/;
						next if  $current_name =~ /subdivision/;
						next if  $current_name =~ /subgen/;
						next if  $current_name =~ /subg\./;
						next if  $current_name =~ /Family/;
					$out_line .= "$current_name;";
				}												
		}
		chop $out_line; # get rid of extra semicolon at end
		print "$out_line\n";
	}
	else
	{
		unless ($line =~ /^\s*$/) 
		{
			print "$line Unknown\n";
			#print STDERR "failed on $orig_line\n";
		}
	}
}

$sth->finish();
$dbh->disconnect();


###########################
# SUBROUTINES
###########################

sub read_config_file
{
    my ($filename) = @_;
    open (INPUT, $filename)  or die "can't open DarkHorse config file $filename\n$!\n";
	my %dh_config = ();
	while (<INPUT>)
	{
		next if ($_ =~ /^\s+$/);
		next if ($_ =~ /^#/);
		if ($_ =~ /\[(.+)\]=\s*(.+)/)
		{
			$dh_config{$1}=$2;		
		}			
	}
	close INPUT;			
    return %dh_config;
}

sub nodes(){
	my($tax_id,%Tax)=@_;
	$recursion_count++;
		
	$sql2 = "select parent_tax_id, rank from nodes where tax_id=$tax_id";
    $sth2 = $dbh->prepare($sql2)
            or dieStackTrace("Cannot prepare SELECT '$sql':\n<b>$DBI::errstr</b>");
    $sth2->execute()
        or dieStackTrace("Cannot execute SELECT '$sql':\n<b>$DBI::errstr</b>");
    %query2 = %{$sth2->fetchrow_hashref};
	$parental_tax_id=$query2{parent_tax_id};
	$Tax{$recursion_count}=$parental_tax_id;	
	
# end recursion when get to root level ($recursion_max prevents runaway)
	if($parental_tax_id == 1 || $recursion_count > $recursion_max)  
	{
		$recursion_count = 0;
		$sth2->finish();
		return %Tax;
	}
	&nodes($parental_tax_id,%Tax);	
}

__END__
