#!/usr/local/bin/perl
# install.pl 
# Sheila Podell 
# April 1, 2009

# $Id: install.pl 161 2009-12-14 23:50:18Z spodell $

# Tests availability of essential pre-requisites for running DarkHorse
# Generates customized DarkHorse config file based on template and user input
# Automates process of installing MySQL databases required to run DarkHorse

use warnings; 
use strict;
use Getopt::Long;
use Cwd;
use File::Basename;
use File::Spec;
use DBI;

# global input:
    my ($program_directory, $input_directory,  $help);
    my ($db, $sth, $sql, $overwrite, $update, $result);
    my $config_template = "templates/config_template";
    my $config_filename = "$config_template";
    
    my $USAGE;
	my $message = qq(
  Usage: $0  -p program directory path 
  Optional parameters: 
    -c configuration file
    -i input data directory
    -o allow database overwrites
    -u allow database updates)."\n\n";
    
	GetOptions( 
			    "p=s" => \$program_directory,
			    "i=s" => \$input_directory,
			    "c=s" => \$config_filename,   # pre-existing or manually edited file
			    'o'   => \$overwrite,		
			    'u'   => \$update,
				);
	 &check_args();
	
# Get default parameters from template, then generate customized config file based on user input    
    my %config_params = &read_config_file("$program_directory/$config_template");
    my $new_filename = &new_filename();
    if ($config_filename ne "$config_template" && -s "$program_directory/$config_filename")
    {
        print "  Using config file $config_filename\n";
    }     
    else
    {
        print "  Creating customized configuration file: $new_filename\n";
        $config_params{program_directory} = $program_directory;
        &write_config_file(\%config_params, $new_filename);
        $config_filename = $new_filename;
    }
  
  # revise parameters interactively if necessary
     %config_params= &read_config_file($config_filename);
     &user_input(\%config_params);  
     
# Install downloaded NCBI taxonomy database tables in local MySQL database
    my $install_script_dir = "$program_directory/bin/installation_scripts";
    my $database = $config_params{db_name};
    my $names = $config_params{names_dmp_path};
    my $nodes = $config_params{nodes_dmp_path};
    my $gi_tax_id_index = $config_params{gi_taxid_prot_dmp_path};
    my $nr_fastapath = $config_params{genbank_nr_fasta_path};
    
    print STDERR "Loading taxonomy names table into database $database\n";
    `$install_script_dir/load_taxonomy_names.pl -d $database -c $config_filename -i $names`;
    print STDERR "Loading taxonomy nodes table into database $database\n";
   `$install_script_dir/load_taxonomy_nodes.pl -d $database -c $config_filename -i $nodes`;
    print STDERR "Loading gi_taxid_prot table into database $database\n";
    #print STDERR "NOTE: This is a really big table - typical number of lines is 3X Genbank nr sequences\n";
   `$install_script_dir/load_gi_taxid_prot_table.pl -c $config_filename `;
    
# Get list of all species names in this version of Genbank nr
    print STDERR "Getting species list for Genbank nr \n";    
    `$install_script_dir/get_nr_species_list.pl -c $config_filename > $input_directory/species_list`;
    unless (-s "$input_directory/species_list")
	{
		print STDERR "ERROR: Creation of file $input_directory/species_list unsuccessful\n";
		exit(0);
	}
  
# pre-calculate lineages by recursively traversing nodes from NCBI taxonomy database
    print STDERR "Pre-calculating lineages for Genbank nr species\n";
    print STDERR "  (NOTE: this step may take 30 minutes or more)\n";      
   `$install_script_dir/recursive_taxonomy.pl -d $database -c $config_filename -i $input_directory/species_list > $input_directory/nr_species_lineages.tab`;
   unless (-s "$input_directory/nr_species_lineages.tab")
 	{
 		print STDERR "ERROR: Creation of file $input_directory/nr_species_lineages.tab unsuccessful\n";
 		exit(0);
 	}

# load precalc_lineages table
     print STDERR "Loading precalc_lineages table into database $database\n";  
    `$install_script_dir/load_precalc_lineages.pl -d $database -c $config_filename -i $input_directory/nr_species_lineages.tab`;     
 
 # get sequence info from Genbank nr fasta file (annotation, sequence length)  
    # takes ~ 3 minutes
	 print STDERR "Collecting annotations and sequence lengths from Genbank nr \n"; 
	   print STDERR "  (NOTE: this step may take 20 minutes or more)\n"; 
	 `$install_script_dir/get_nr_seq_info.pl $nr_fastapath > $input_directory/nr.info`;	 
	unless (-s "$input_directory/nr.info")
 	{
 		print STDERR "ERROR: Creation of file $input_directory/nr.info unsuccessful\n";
 		exit(0);
 	}

# combine sequence info with taxonomy info, use to load db_seq_ids table
	 print STDERR "Combining taxonomy info with sequence data\n";
	`$install_script_dir/get_db_seq_table_input.pl -c $config_filename -i $input_directory/nr.info > $input_directory/nr_all_ids.tab`;	
	unless (-s "$input_directory/nr_all_ids.tab")
	{
 		print STDERR "ERROR: Creation of file $input_directory/nr_all_ids.tab unsuccessful\n";
 		exit(0);
 	}

# load rest of tables	 
	 print STDERR "Loading table db_seq_ids into database $database\n";
	 `$install_script_dir/load_db_seq_ids_table.pl -d $database -c $config_filename -i $input_directory/nr_all_ids.tab`;
    
  # account for missing lineages, so they don't cause problems later
    print STDERR "Creating master lineage index\n";
    `$install_script_dir/create_missing_lineage_list.pl -d $database -c $config_filename -a > $input_directory/all_lineages.tab`;
    unless (-s "$input_directory/all_lineages.tab")
 	{
 		print STDERR "ERROR: Creation of file $input_directory/all_lineages.tab unsuccessful\n";
 		exit(0);
 	}
   
    print STDERR "Loading table lineage_index into database $database\n";     
    `$install_script_dir/load_lineage_index_table.pl  -d $database -c $config_filename -i $input_directory/all_lineages.tab`;

# user feedback
	print STDERR "Finished loading MySQL tables\n";

#########################################################
# SUBROUTINES
#########################################################
sub check_args
{
 	if($USAGE || !$program_directory) 
	{
		print STDERR $message;
		exit(0);
	} 
	unless (-s "$program_directory")
	{
		print STDERR "  Couldn't find program directory $program_directory \n";
		exit(0);
	}
	unless (-s "$program_directory/$config_template")
	{
		print STDERR "  Couldn't find config template $program_directory/$config_template \n";
		exit(0);		
	}
	# check for required program files
        my @program_list = (
                            "load_taxonomy_names.pl",
                            "load_taxonomy_nodes.pl",
                            "load_gi_taxid_prot_table.pl",
                            "get_nr_species_list.pl",
                            "recursive_taxonomy.pl",
                            "load_precalc_lineages.pl",
                            "get_nr_seq_info.pl",
                            "get_db_seq_table_input.pl",
                            "load_db_seq_ids_table.pl",                           
                            "create_missing_lineage_list.pl",
                            "load_lineage_index_table.pl",                            
                            );
         
         foreach (@program_list)
         {
           my $program =  "$program_directory/bin/installation_scripts/$_";
            unless (-s $program && -x $program)
            {
                print STDERR "Couldn't find program $_\n";
                exit (0);
            }         
         }
	# setup data input directory
        if (defined $input_directory && -s "$input_directory")
        {
            return;
        }
            
        $input_directory = "$program_directory/input_data";
        print STDERR "  Using data input directory $program_directory/input_data\n";	
        
        unless (-s 	$input_directory)
        {
            mkdir "$program_directory/input_data", 0755 or die "Cannot make input_data directory $program_directory/input_data\n: $!"; 		
        }	
}

sub read_config_file
{
    my ($filename) = @_;
    open (INPUT, $filename)  or die "can't open DarkHorse config file $filename\n$!\n";
	my %dh_config = ();
	while (<INPUT>)
	{
		next if ($_ =~ /^\s+$/);
		next if ($_ =~ /^#/);
		if ($_ =~ /\[(.+)\]=\s*(.+)/)
		{
			$dh_config{$1}=$2;		
		}			
	}
	close INPUT;			
    return %dh_config;
}

sub write_config_file
{  
	my ($hashref, $newfile) = @_;
	my %parameters = %$hashref;
	
	my $date = `date`;
	chomp $date;
	    
	open (CONFIG,  ">$program_directory/$newfile") or die "couldn't create new configuration file\n$!\n\n";			
	print CONFIG "# Configuration file for DarkHorse $newfile, generated by dh_installation.pl\n";
	print CONFIG "# $date\n\n";
				
	foreach (keys %parameters)
	{
	    print CONFIG "[$_]=$parameters{$_}\n";
	
	}		
    print CONFIG "\n";		 
	close CONFIG;
}

sub new_filename
{
    my $newname = "";
    # don't overwrite old file
        if (-s "$program_directory/darkhorse.cfg")
        {	    
            `cp $program_directory/darkhorse.cfg $program_directory/darkhorse_old.cfg`;	
        }
	
	# put a unique, sequential number on the new filename
        my $basename = "darkhorse";
        my $counter = 1;
        do{
            $newname = "$basename$counter".".cfg";
            $counter++;
        }while (-s "$program_directory/$newname" );
        
	return $newname;
}	
	
sub user_input
{    
    my $changes = 0;
   
    
    # confirm location of non-executable file paths    
        my @file_list = ( "genbank_nr_fasta_path",
                          "names_dmp_path",
                          "nodes_dmp_path",
                          "gi_taxid_prot_dmp_path",
                         );
        print "\n  Please enter corrected location for the following paths, if required\n";
        foreach (@file_list)
        {
            my $path = $config_params{$_};
            $path = " " unless (defined $path);
            
            print "  $_ ($path): \n";
            my $tmp = <STDIN>;
            chomp $tmp;
            if (length $tmp>1)
            {
                $config_params{$_} = $tmp;
                $changes++;
            }
            unless (-s "$config_params{$_}")
            {
                print  "  Can't find $_ in specified location:\n  $config_params{$_}\n";
                print  "  DarkHorse configuration cannot be completed without this file.\n";
                unlink "$program_directory/$new_filename";
                exit (0);      
            }        
        }

    # confirm mysql connection host, username, password, and database
        my @db_params = ("db_name","db_host","db_user","db_user_password");
        print "  Please enter corrected connection information for MySQL database, if necessary.\n";        
        foreach (@db_params)
        {           
            print "  $_ ($config_params{$_}):\n";
            my $tmp = <STDIN>;
            chomp $tmp;
            if (length $tmp>1)
            {
                $config_params{$_} = $tmp;
                $changes++;
            }                   
        }
        my $result = &check_db_connection(\%config_params);
        unless ($result ==1)
        {
            print "\n  Unable to access database using the following parameters:\n";
             foreach (@db_params)
            {           
                print "    $_ ($config_params{$_})\n";    
            }
            print "  If database $config_params{db_name} does not already exist, you must create it manually.\n";
            exit(0);
        }
        print "Database connection OK\n";       
        
    # check format for max_lines_per_packet, minimum number of lineage terms, min_align_coverage
        unless ($config_params{max_lines_per_packet} =~ /^\d+$/)
        {
            print STDERR "Illegal value for max_lines_per_packet:$config_params{max_lines_per_packet}\n";
            exit(0);	
        }
        unless ($config_params{min_lineage_terms} =~ /^\d+$/)
        {
            print STDERR "Illegal value for min_lineage_terms:($config_params{min_lineage_terms}\n";
            exit(0);	
        }
        unless ($config_params{min_align_coverage} >=0 && $config_params{min_align_coverage} <= 1)
        {
            print STDERR "Illegal value for min_align_coverage:($config_params{min_align_coverage}\n";
            exit(0);	
        }
        
        if ($changes>0)
        {
            &write_config_file(\%config_params, $new_filename);
        }
}

sub check_db_connection
{
    my ($hashref) = @_;
	my %dh_config = %$hashref;
	my $found_db = 0;
	
	my $db_name = $dh_config{db_name} || "not found";
	my $db_program = $dh_config{db_program} || "not found";
	my $db_host = $dh_config{db_host} || "not found";
	my $db_user = $dh_config{db_user} || "not found";
	my $db_user_password = $dh_config{db_user_password} || "not found";
    
    my $db_path = "DBI:$db_program:$db_name:$db_host:";
    
	my $dbh = DBI->connect($db_path, $db_user, $db_user_password);
	unless ($dbh)
	{
	    return $found_db;	    
	}

	my $sql = "show databases";
	my $sth = $dbh->prepare($sql)
    	or dieStackTrace("Cannot prepare SELECT '$sql':\n<b>$DBI::errstr</b>");
	$sth->execute()
    	or dieStackTrace("Cannot execute SELECT '$sql':\n<b>$DBI::errstr</b>");
	my @row = ();
	my $element = ();
	while (@row = $sth->fetchrow_array)
	{
		foreach $element (@row)
		{
			if ($element =~/$db_name/)
			{
				$found_db = 1;
				last;
			}
		}
	}	
	
	$dbh->disconnect();
	return $found_db;
}

sub missing_file_msg
{
       print  "  This program is required for DarkHorse configuration.\n";
                print  qq(  It can be downloaded from ftp://ftp.ncbi.nih.gov/blast/)."\n";
                unlink "$program_directory/$new_filename";
                exit (0);
}


  
