#!/usr/local/bin/perl
# test.pl 
# Sheila Podell 
# August 14, 2007

use warnings; 
use strict;
use Getopt::Long;
use DBI;

# Tests ability to run DarkHorse using standard test set
# Output to directory called "test_results"
# (Intentionally) Overwrites previous results if directory has same name

# global parameters
    my $config_filename;    
    my $USAGE;
	my $message = qq( Usage: $0  -c configuration file)."\n\n";    
	GetOptions("c=s" => \$config_filename);

# Get data from config file   
    if($USAGE || ! $config_filename) 
    {
        print STDERR $message;
        exit(0);
    }         
     unless (-s $config_filename)
     {
        print STDERR "Couldn't read config file $config_filename\n";
        exit(0);
     }
     
     print STDERR "Testing DarkHorse using config file: $config_filename\n";
     my %config_params = &read_config_file("$config_filename");
    &check_args();

# make sure test files are accessible 
   my $program_directory = $config_params{program_directory};
   my $test_data_dir = "$program_directory/test_data";
   my $test_fasta = "$test_data_dir/test.fasta";
   my $test_blast = "$test_data_dir/test.blast";
   my $test_exclude = "$test_data_dir/test_exclude_list";
   my $test_match_ids = "$test_data_dir/test_match_ids";
   my $test_matches_tab = "$test_data_dir/test_matches.tab";
   my $test_lineages_tab = "$test_data_dir/test_lineages.tab";
   
   my @file_list = ("$test_fasta","$test_blast","$test_exclude","$test_match_ids", 
                    "$test_matches_tab", "$test_lineages_tab");         
   foreach (@file_list)
   {
      unless (-s $_)
      {
          print STDERR "Couldn't find required test_data file $_\n";
          exit (0);
      }         
   }	
   
# make sure database is accessible, test file matches are loaded       
   open (INFILE, "$test_match_ids") or die "couldn't open $test_match_ids\n$!\n";
   chomp (my @test_match_list = <INFILE>);
   close INFILE;
    
   my $result = &check_db_connection(\%config_params, \@test_match_list);    

# Run DarkHorse on test files
    print STDERR "Running DarkHorse program test\n";
    my $blast_filter = 0.7; 

    `$program_directory/bin/darkhorse.pl -t $test_blast  -e $test_exclude -g $test_fasta -c $config_filename -b $blast_filter`;
    my @list = `ls -alt |grep filt70 | head -3 `;
    foreach my $line (@list)
    {
        chomp $line;
        if ($line =~ /.*\s(.+filt70.*)$/)
        {           
           my $filename = $1;
           #print STDERR "  deleting $filename\n";
            unlink $filename;
        }         
    }

# over-write old test_results
    if (-s "test_results")
    {
        &rmdir_recursive("test_results");
    }
   my $result_dir = "";
    my $result_check = `ls -alt |grep calcs |head -1`;
    chomp $result_check;
    if (defined $result_check && $result_check =~ /.*\s(calcs_.*)$/)
    {
        $result_dir = "$1";
        `mv $result_dir/ test_results`;
        $result_dir = "test_results";
        print STDERR "Results successfully written to $result_dir.\n";
    }   

#########################################################
# SUBROUTINES
#########################################################

sub read_config_file
{
    my ($filename) = @_;
    open (INPUT, $filename)  or die "can't open DarkHorse config file $filename\n$!\n";
	my %dh_config = ();
	while (<INPUT>)
	{
		next if ($_ =~ /^\s+$/);
		next if ($_ =~ /^#/);
		if ($_ =~ /\[(.+)\]=\s*(.+)/)
		{
			$dh_config{$1}=$2;		
		}			
	}
	close INPUT;			
    return %dh_config;
}

sub check_args
{
	my $program_directory = $config_params{program_directory};
	unless (-s "$program_directory")
	{
		print STDERR "  Couldn't find program directory $program_directory \n";
		exit(0);
	}
	# check for required program files
        my @program_list = ("select_tab_first.pl",
                            "select_tab_equivalent.pl",
                            "getseq_multiple.pl",
                            "get_seq_info.pl",
                            "calc_lineage_probability.pl",
                            "filter_blast.pl",);         
         foreach (@program_list)
         {
           my $program =  "$program_directory/bin/accessory_scripts/$_";
            unless (-s $program && -x $program)
            {
                print STDERR "Couldn't find program $_\n";
                exit (0);
            }
         
         }	
}

sub check_db_connection
{
    my ($hashref, $arrayref) = @_;
	my %dh_config = %$hashref;
	my @matchlist = @$arrayref;
	my $num_queries = scalar @matchlist;
	my $found_db = 0;
		
	my $db_name = $dh_config{db_name} || "not found";
	my $db_program = $dh_config{db_program} || "not found";
	my $db_host = $dh_config{db_host} || "not found";
	my $db_user = $dh_config{db_user} || "not found";
	my $db_user_password = $dh_config{db_user_password} || "not found";
    
    my $db_path = "DBI:$db_program:$db_name:$db_host:";
    
	my $dbh = DBI->connect($db_path, $db_user, $db_user_password);
	unless ($dbh)
	{
	   print "\n  ERROR - unable to access database specified in configuration file.\n";
        exit(0);	    
	}

	my $sql = "show databases";
	my $sth = $dbh->prepare($sql)
    	or dieStackTrace("Cannot prepare SELECT '$sql':\n<b>$DBI::errstr</b>");
	$sth->execute()
    	or dieStackTrace("Cannot execute SELECT '$sql':\n<b>$DBI::errstr</b>");
	my @row = ();
	my $element = ();
	while (@row = $sth->fetchrow_array)
	{
		foreach $element (@row)
		{
			if ($element =~/$db_name/)
			{
				$found_db = 1;
				last;
			}
		}
	}
	unless ($found_db ==1)
    {
        print "\n  ERROR - unable to access database specified in configuration file.\n";
        exit(0);
    }
    print "Database connection OK\n";
    
 
	$sql = "select full_id from db_seq_ids where full_id in (";
	foreach (@matchlist)
	{
		$sql .= "'$_',";
	}
	$sql = substr($sql, 0, -1); # get rid of extra newline and comma, if present
	$sql .= ");\n";
	
	$sth = $dbh->prepare($sql)
    	or dieStackTrace("Cannot prepare SELECT '$sql':\n<b>$DBI::errstr</b>");
	$sth->execute()
    	or dieStackTrace("Cannot execute SELECT '$sql':\n<b>$DBI::errstr</b>");
    my $num_found = $sth->rows;
    $sth->finish();
    $dbh->disconnect();
    
    unless ($num_found == $num_queries)
    {
        print STDERR "Test_matchlist not yet loaded in database\n";
        print STDERR "Attempting to load them now\n"; 
        my $database = $config_params{db_name};       
        my $load_script_1 = "$program_directory/bin/installation_scripts/load_db_seq_ids_table.pl";
        my $load_script_2 = "$program_directory/bin/installation_scripts/load_lineage_index_table.pl";
        `$load_script_1 -d $database -c $config_filename -i $test_matches_tab -u`;
        `$load_script_2  -d $database -c $config_filename -i $test_lineages_tab -u `;
    }
	return $found_db;
}

sub rmdir_recursive
{
     my ($dirname) = @_;
     if (-s "$dirname")  
    {
        opendir (THISDIR, "$dirname") or die "can't open directory listing for $dirname";
	    my @file_list = readdir THISDIR;
	    closedir THISDIR;
	    foreach my $file (@file_list)
	    {
             next if $file =~ /^\.{1,2}$/;
	        unlink "$dirname/$file";
	    }
	    rmdir  $dirname;
    }
}